' Transpiled on 28-09-2022 21:45:17

' Copyright (c) 2022 Thomas Hugo Williams
' License MIT <https://opensource.org/licenses/MIT>
' For MMBasic 5.07.05

Option Base 0
Option Default None
Option Explicit On

' Option LcdPanel NoConsole

' BEGIN:     #Include "PicoGAME.inc" -------------------------------------------
' Copyright (c) 2022 Thomas Hugo Williams
' License MIT <https://opensource.org/licenses/MIT>
' For PicoMite VGA, MMBasic 5.07.05

' Note 1:
'   This code is written for readability, not for performance.
'   If necessary it can be made faster by:
'    - inlining constants to avoid overhead of "variable" lookup,
'      this will also reduce the amount of RAM required.
'    - unrolling FOR loops.
'    - shortening variable names.
'    - using different functions to read ports A and B so that pin numbers
'      can be hard-coded.
'    - in extremis inlining the xxx_read%() functions to avoid the overhead
'      of a subroutine call.
'
' Note 2:
'   Variables with CAPITALISED names are constants, this includes arrays
'   which have to be declared with DIM and where constant-ness is not enforced
'   by the MMBasic interpreter.

If MM.Info(Option Base) <> 0 Then Error "OPTION BASE 0 required"

' Button values as returned by xxx_read%() functions.
Const NES_A%      = 1  ' FIRE on joystick and "B" on SNES.
Const NES_B%      = 2  ' "Y" on SNES.
Const NES_SELECT% = 4
Const NES_START%  = 8
Const NES_UP%     = 16
Const NES_DOWN%   = 32
Const NES_LEFT%   = 64
Const NES_RIGHT%  = 128
Const SNES_A%     = 256
Const SNES_X%     = 512
Const SNES_LB%    = 1204
Const SNES_RB%    = 2048

If Instr(MM.Device$, "PicoMite") > 0 Then

' Pin mappings for ATARI 1-button joystick.
Dim ATARI_FIRE_PIN%(1)  = ( MM.Info(PinNo GP14), MM.Info(PinNo GP28) )
Dim ATARI_UP_PIN%(1)    = ( MM.Info(PinNo  GP0), MM.Info(PinNo GP28) )
Dim ATARI_DOWN_PIN%(1)  = ( MM.Info(PinNo  GP1), MM.Info(PinNo  GP1) )
Dim ATARI_LEFT_PIN%(1)  = ( MM.Info(PinNo  GP2), MM.Info(PinNo  GP2) )
Dim ATARI_RIGHT_PIN%(1) = ( MM.Info(PinNo  GP3), MM.Info(PinNo  GP3) )

' Pin mappings for NES/SNES gamepad.
Dim NES_CLOCK_PIN%(1) = ( MM.Info(PinNo  GP3), MM.Info(PinNo GP22) )
Dim NES_DATA_PIN%(1)  = ( MM.Info(PinNo  GP1), MM.Info(PinNo  GP4) )
Dim NES_LATCH_PIN%(1) = ( MM.Info(PinNo  GP2), MM.Info(PinNo  GP5) )

EndIf

' The NES standard specifies a 12 micro-second pulse, but all the controllers
' I've tested work with 1 micro-second, and possibly less.
Const NES_PULSE_DURATION! = 0.001 ' 1 micro-second

Dim CONTROLLER_IDS$(3) Length 8 = ( "keyboard", "atari", "nes", "snes" )
Dim CONTROLLER_NAMES$(3) Length 15 = ( "Keyboard", "Atari Joystick", "NES Gamepad" , "SNES Gamepad" )

' +----------------------------------------------------------------------------+
' |                         READING GAME CONTROLLERS                           |
' +----------------------------------------------------------------------------+

' Waits a specified duration for the user to press START on a (S)NES gamepad
' connected to Port A, or FIRE on an ATARI joystick connected to Port A or
' SPACE on the keyboard.
'
' Note that this function deliberately does not clear the keyboard buffer before
' checking to see if SPACE has been pressed; if desired this can be done by
' calling keyboard_init().
'
' @param   duration%  duration in milliseconds; if 0 then indefinite.
' @return             "nes" if the (S)NES START button is pressed,
'                     or "atari" if the ATARI FIRE button is pressed,
'                     or "keyboard" if the SPACE key is pressed,
'                     or "" if none of the above before the duration expires.
Function poll_for_controller$(duration%)
  Local expires% = Choice(duration%, Timer + duration%, &h7FFFFFFFFFFFFFFF)

  ' Short-circuit if we are not on a PicoMite.
  If Instr(MM.Device$, "PicoMite") = 0 Then
    Do While poll_for_controller$ = "" And Timer < expires%)
      poll_for_controller$ = Choice(keyboard_read%() = NES_A%, "keyboard", "")
    Loop
    Exit Function
  EndIf

  Do While poll_for_controller$ = "" And Timer < expires%)
    If poll_for_controller$ = "" Then
      nes_a_init()
      poll_for_controller$ = Choice(nes_a_read%() = NES_START%, "nes_a", "")
    EndIf
    If poll_for_controller$ = "" Then
      atari_a_init()
      poll_for_controller$ = Choice(atari_a_read%() = NES_A%, "atari_a", "")
    EndIf
    If poll_for_controller$ = "" Then
      poll_for_controller$ = Choice(keyboard_read%() = NES_A%, "keyboard", "")
    EndIf
  Loop
End Function

' Initialises port A connected to an ATARI 1-button joystick.
Sub atari_a_init()
  SetPin ATARI_UP_PIN%(0), DIn
  SetPin ATARI_DOWN_PIN%(0), DIn
  SetPin ATARI_LEFT_PIN%(0), DIn
  SetPin ATARI_RIGHT_PIN%(0), DIn
  SetPin ATARI_FIRE_PIN%(0), DIn
End Sub

' Initialises port B connected to an ATARI 1-button joystick.
Sub atari_b_init()
  SetPin ATARI_UP_PIN%(1), DIn
  SetPin ATARI_DOWN_PIN%(1), DIn
  SetPin ATARI_LEFT_PIN%(1), DIn
  SetPin ATARI_RIGHT_PIN%(1), DIn
  SetPin ATARI_FIRE_PIN%(1), DIn
End Sub

' Reads port A connected to an ATARI 1-button joystick.
'
' @return  bitmap corresponding to current controller state:
'            7 6 5 4 3 2 1 0
'            | | | | | | | |
'            | | | | | | | +-- Fire Button
'            | | | | | | +---- unused
'            | | | | | +------ unused
'            | | | | +-------- unused
'            | | | +---------- Up
'            | | +------------ Down
'            | +-------------- Left
'            +---------------- Right
Function atari_a_read%()
  Inc atari_a_read%, Not Pin(ATARI_FIRE_PIN%(0))
  Inc atari_a_read%, Not Pin(ATARI_UP_PIN%(0)) * 16
  Inc atari_a_read%, Not Pin(ATARI_DOWN_PIN%(0)) * 32
  Inc atari_a_read%, Not Pin(ATARI_LEFT_PIN%(0)) * 64
  Inc atari_a_read%, Not Pin(ATARI_RIGHT_PIN%(0)) * 128
End Function

' Reads port B connected to an ATARI 1-button joystick.
Function atari_b_read%()
  Inc atari_b_read%, Not Pin(ATARI_FIRE_PIN%(1))
  Inc atari_b_read%, Not Pin(ATARI_UP_PIN%(1)) * 16
  Inc atari_b_read%, Not Pin(ATARI_DOWN_PIN%(1)) * 32
  Inc atari_b_read%, Not Pin(ATARI_LEFT_PIN%(1)) * 64
  Inc atari_b_read%, Not Pin(ATARI_RIGHT_PIN%(1)) * 128
End Function

' Initialises the keyboard by clearing the input buffer.
Sub keyboard_init(idx%)
  Do While Inkey$ <> "" : Loop
End Sub

' Reads the keyboard as if it were a controller.
'
' Note that the PicoMite has no KEYDOWN function so we are limited to
' reading a single keypress from the input buffer and cannot handle multiple
' simultaneous keys or properly handle a key being pressed and not released.
'
' @return  bitmap corresponding to key read:
'            7 6 5 4 3 2 1 0
'            | | | | | | | |
'            | | | | | | | +-- Space
'            | | | | | | +---- unused
'            | | | | | +------ unused
'            | | | | +-------- unused
'            | | | +---------- Up
'            | | +------------ Down
'            | +-------------- Left
'            +---------------- Right
Function keyboard_read%()
  Select Case Asc(Inkey$)
    Case 32  : keyboard_read% = NES_A%
    Case 128 : keyboard_read% = NES_UP%
    Case 129 : keyboard_read% = NES_DOWN%
    Case 130 : keyboard_read% = NES_LEFT%
    Case 131 : keyboard_read% = NES_RIGHT%
  End Select
End Function

' Initialises port A connected to a SNES gamepad.
Sub snes_a_init()
  nes_a_init()
End Sub

' Initialises port B connected to a SNES gamepad.
Sub snes_b_init()
  nes_b_init()
End Sub

' Reads port A connected to a SNES gamepad.
'
' @return  bitmap corresponding to current controller state:
'            11 10 9 8 7 6 5 4 3 2 1 0
'            |  |  | | | | | | | | | |
'            |  |  | | | | | | | | | +-- Button B
'            |  |  | | | | | | | | +---- Button Y
'            |  |  | | | | | | | +------ Select
'            |  |  | | | | | | +-------- Start
'            |  |  | | | | | +---------- Up
'            |  |  | | | | +------------ Down
'            |  |  | | | +-------------- Left
'            |  |  | | +---------------- Right
'            |  |  | +------------------ Button A
'            |  |  +-------------------- Button X
'            |  +----------------------- Left Bumper
'            +-------------------------- Right Bumper
Function snes_a_read%()
  Local i%
  Pulse NES_LATCH_PIN%(0), NES_PULSE_DURATION!
  For i% = 0 To 11
    Inc snes_a_read%, Not Pin(NES_DATA_PIN%(0)) * 2^i%
    Pulse NES_CLOCK_PIN%(0), NES_PULSE_DURATION!
  Next
End Function

' Reads port B connected to a SNES gamepad.
Function snes_b_read%()
  Local i%
  Pulse NES_LATCH_PIN%(1), NES_PULSE_DURATION!
  For i% = 0 To 11
    Inc snes_b_read%, Not Pin(NES_DATA_PIN%(1)) * 2^i%
    Pulse NES_CLOCK_PIN%(1), NES_PULSE_DURATION!
  Next
End Function

' Initialises port A connected to a NES gamepad.
Sub nes_a_init()
  SetPin NES_LATCH_PIN%(0), Dout
  SetPin NES_CLOCK_PIN%(0), Dout
  SetPin NES_DATA_PIN%(0), Din
  Pin(NES_LATCH_PIN%(0)) = 0
  Pin(NES_CLOCK_PIN%(0)) = 0
End Sub

' Initialises port B connected to a NES gamepad.
Sub nes_b_init()
  SetPin NES_LATCH_PIN%(1), Dout
  SetPin NES_CLOCK_PIN%(1), Dout
  SetPin NES_DATA_PIN%(1), Din
  Pin(NES_LATCH_PIN%(1)) = 0
  Pin(NES_CLOCK_PIN%(1)) = 0
End Sub

' Reads port A connected to a NES gamepad.
'
' @return  bitmap corresponding to current controller state:
'            7 6 5 4 3 2 1 0
'            | | | | | | | |
'            | | | | | | | +-- Button A
'            | | | | | | +---- Button B
'            | | | | | +------ Select
'            | | | | +-------- Start
'            | | | +---------- Up
'            | | +------------ Down
'            | +-------------- Left
'            +---------------- Right
'
' Note that in theory it should not be necessary to pulse the clock pin after
' reading bit 7 (Right), but in practice there are some clone NES controllers
' which return the wrong value on subsequent reads of bit 7 if it not pulsed.
Function nes_a_read%()
  Local i%
  Pulse NES_LATCH_PIN%(0), NES_PULSE_DURATION!
  For i% = 0 To 7
    Inc nes_a_read%, Not Pin(NES_DATA_PIN%(0)) * 2^i%
    Pulse NES_CLOCK_PIN%(0), NES_PULSE_DURATION!
  Next
End Function

' Reads port B connected to a NES gamepad.
Function nes_b_read%()
  Local i%
  Pulse NES_LATCH_PIN%(1), NES_PULSE_DURATION!
  For i% = 0 To 7
    Inc nes_b_read%, Not Pin(NES_DATA_PIN%(1)) * 2^i%
    Pulse NES_CLOCK_PIN%(1), NES_PULSE_DURATION!
  Next
End Function

' Gets a string representation of bits read from a controller.
'
' @param id$    controller Id from CONTROLLER_IDS$().
' @param bits%  controller state returned by xxx_read%() function.
Function ctrl_bits_to_string$(id$, bits%)
  ' These are the names of the buttons on a SNES controller;
  ' we adjust them in the SELECT CASE below for NES/ATARI names.
  Static BUTTONS$(11) = ("B","Y","Select","Start","Up","Down","Left","Right","A","X","L","R")

  If bits% = 0 Then Exit Function

  ctrl_bits_to_string$ = Str$(bits%) + " = "
  Local count%, i%, s$
  For i% = 0 To Bound(BUTTONS$(), 1)
    If bits% And 2^i% Then
      s$ = BUTTONS$(i%)
      Select Case id$
        Case "atari"
          If s$ = "B" Then s$ = "Fire"
        Case "nes"
          If s$ = "B" Then s$ = "A"
          If s$ = "Y" Then s$ = "B"
        Case "keyboard", "snes"
          ' No extra manipulation.
        Case Else
          Error "Unknown controller: " + id$
      End Select
      If count% > 0 Then Inc ctrl_bits_to_string$, ", "
      Inc ctrl_bits_to_string$, s$
      Inc count%
    EndIf
  Next
End Function
' END:       #Include "PicoGAME.inc" -------------------------------------------
' BEGIN:     #Include "highscore.inc" ------------------------------------------
' Copyright (c) 2022 Thomas Hugo Williams
' License MIT <https://opensource.org/licenses/MIT>
' For MMBasic 5.07.05

Const highscore.MAX_NAME_LEN% = 8

Dim highscore.values$(9) Length highscore.MAX_NAME_LEN% + 7

Sub highscore.load(filename$, defaults$())
  ' It is convenient to always fill with the defaults
  ' even if we end up overwriting them immediately.
  If Bound(highscore.values$(), 1) <> Bound(defaults$()) Then
    Error "Invalid number of high-score defaults"
  EndIf
  Local i%
  For i% = 0 To Bound(highscore.values$())
    highscore.values$(i%) = defaults$(i%)
  Next

  ' If there is no SD Card then do not load high-scores.
  Local has_sdcard% = UCase$(MM.Info(SdCard)) = "READY"
  If Not has_sdcard% Then Exit Sub

  ' Check filename$ is absolute.
  Local ok% = 0
  Select Case Left$(filename$, 1)
    Case "/", "\" : ok% = 1
    Case "A", "a" : If Mid$(filename$, 2, 1) = ":" Then ok% = 1
  End Select
  If Not ok% Then Error "Expected absolute file path '" + filename$ + "'"

  ' Create any parent directories for filename$ if they do not exist.
  Local ch$, parent$
  For i% = 1 To Len(filename$)
    ch$ = Mid$(filename$, i%, 1)
    Select Case ch$
      Case "/", "\"
        If parent$ = "" Or UCase$(parent$) = "A:" Then Continue For
        Select Case MM.Info(FileSize parent$)
          Case -2   : ' Is a directory, do nothing.
          Case -1   : Mkdir parent$
          Case Else : Error "Expected directory but found file '" + parent$ + "'"
        End Select
      Case Else
        Inc parent$, ch$
    End Select
  Next

  ' Check file exists and is not a directory.
  Select Case MM.Info(FileSize filename$)
    Case -2: Error "Expected file but found directory '" + parent$ + "'"
    Case -1: Exit Sub ' File does not exist.
  End Select

  ' Read the file.
  Open filename$ For Input As #1
  For i% = 0 To Bound(highscore.values$())
    Line Input #1, highscore.values$(i%)
    ok% = ok% And Field$(highscore.values$(i%), 1) <> ""
    ok% = ok% And Field$(highscore.values$(i%), 2) <> ""
    ok% = ok% And Val(Field$(highscore.values$(i%), 2)) > 0
    If Not ok% Then Error "Invalid high-score file '" + filename$ + "'"
  Next
  Close #1
End Sub

' Assumes parent directory of filename$ already exists.
Sub highscore.save(filename$)
  ' If there is no SD Card then do not save high-scores.
  Local has_sdcard% = UCase$(MM.Info(SdCard)) = "READY"
  If Not has_sdcard% Then Exit Sub

  Open filename$ For Output As #1
  Local i%
  For i% = 0 To Bound(highscore.values$())
    Print #1, highscore.values$(i%)
  Next
  Close #1
End Sub

' Shows the high-score table for a specified duration or until the user presses
' START/FIRE/SPACE.
'
' @param duration%  duration in milliseconds; if 0 then indefinite.
' @return           1 if the user pressed button/key,
'                   0 if the duration expired.
Function highscore.show_table%(duration%)
  Const ch$ = Chr$(205)
  Text X_OFFSET%, Y_OFFSET% - 95, ch$ + ch$ + " HIGH SCORES " + ch$ + ch$, "CT", 1, 1, RGB(White)

  Local col_idx%, ctrl$, i%, name$, score$, y%
  Local expires% = Choice(duration%, Timer + duration%, &h7FFFFFFFFFFFFFFF)
  Local colours%(3) = (RGB(Red), RGB(Yellow), RGB(Cyan), RGB(Green))

  keyboard_init()
  Do While Timer < expires% And ctrl$ = ""

    For i% = 0 To Bound(highscore.values$(), 1) + 1
      If i% <= Bound(highscore.values$(), 1) Then
        name$ = str.rpad$(Field$(highscore.values$(i%), 1), highscore.MAX_NAME_LEN%)
        score$ = str.lpad$(Field$(highscore.values$(i%), 2), 5)
        y% = Y_OFFSET% - 75 + 15 * i%
        Text X_OFFSET%, y%, score$ + "  " + name$, "CT", 1, 1, colours%(col_idx%)
      EndIf
      col_idx% = (col_idx% + 1) Mod 4
    Next

    If Not(Instr(MM.Device$, "PicoMite")) Then Page Copy 1 To 0, B

    ctrl$ = poll_for_controller$(200)
  Loop

  highscore.show_table% = ctrl$ <> ""
End Function

' Provides UI for editing an entry in the high-score table.
'
' @param  idx%      index of entry in the highscore.values$ array.
' @param  colour_%  player colour.
' @param  ctrl$     controller for given player.
Sub highscore.edit(idx%, colour_%, ctrl$)
  ' 256-bit map of characters that can be entered into the high-score table
  ' via gamepad or joystick. For each character 0-255 the character is allowed
  ' if the corresponding bit is set.
'  Static ALLOWED_CHARS%(3) = (&h03FF440100000000, &h000000007FFFFFF, &h0000020080001E00, &h8000000000000000)
  Static ALLOWED_CHARS%(3) = (&h03FF440100000000, &h000000007FFFFFF, &h0000020080001E00, &h0)

  Local ch$ = Chr$(205)
  Text X_OFFSET%, Y_OFFSET% - 95, ch$ + ch$ + " HIGH SCORES " + ch$ + ch$, "CT", 1, 1, RGB(White)

  ' Draw current high-score table.
  Local i%, name$, score$, x%, y%
  For i% = 0 To Bound(highscore.values$(), 1)
    name$ = str.rpad$(Field$(highscore.values$(i%), 1), highscore.MAX_NAME_LEN%)
    score$ = str.lpad$(Field$(highscore.values$(i%), 2), 5)
    y% = Y_OFFSET% - 75 + 15 * i%
    Text X_OFFSET%, y%, score$ + "  " + name$, "CT", 1, 1, Choice(i% = idx%, colour_%, RGB(Blue))
  Next

  Local cursor% = 1, key%, p% = 1, t% = Timer, bg%, fg%, play_sound%
  y% = Y_OFFSET% - 75 + 15 * idx%
  name$ = ""

  ' Re-initialise controller as earlier call to poll_for_controller$() will
  ' have reset pins.
  keyboard_init()
  Select Case ctrl$
    Case "nes_a_read%"   : nes_init()
    Case "nes_b_read%"   : nes_init()
    Case "atari_a_read%" : atari_init()
    Case "atari_b_read%" : atari_init()
  End Select

  Local font_width% = Peek(Byte MM.Info(Font Address 1))
  Do
    If Timer > t% + 500 Then
      cursor% = Not cursor%
      t% = Timer
    EndIf

    For i% = 1 To highscore.MAX_NAME_LEN%
      bg% = Choice((i% = p%) And cursor%, colour_%, RGB(Black))
      fg% = Choice((i% = p%) And cursor%, RGB(Black), colour_%)
      ch$ = Choice(i% <= Len(name$), Mid$(name$, i%, 1), " ")
      x% = X_OFFSET% + (i% - 1.5) * font_width%
      Text x%, y%, ch$, , 1, 1, fg%, bg%
    Next

    ' Display green check mark if next right/fire will OK the name.
    Inc x%, 2 * font_width%
    Text x%, y%, Chr$(129), , 1, 1, Choice(p% = highscore.MAX_NAME_LEN%, RGB(Green), RGB(Black))

    If Not(Instr(MM.Device$, "PicoMite")) Then Page Copy 1 To 0, B
    Pause 100

    key% = highscore.get_input%(ctrl$)
    If key% > 0 Then
      play_sound% = 1
      Select Case key%
        Case &h08 ' Backspace
          If p% > 1 Then
            name$ = Left$(name$, p% - 2) + Mid$(name$, p%)
            Inc p%, -1
          EndIf
        Case &h7F ' Delete
          name$ = Left$(name$, p% - 1) + Mid$(name$, p% + 1)
        Case &h0A, &h0D ' LF and CR
          key% = -1 ' So we exit the DO LOOP.
        Case &h20 To &h7E
          Do While p% > Len(name$) : Inc name$, " " : Loop
          name$ = Left$(Left$(name$, p% - 1) + Chr$(key%) + Mid$(name$, p%), highscore.MAX_NAME_LEN%)
          If p% < highscore.MAX_NAME_LEN% Then Inc p%, 1
        Case &h80, &h81 ' Up and down arrows
          i% = Asc(Mid$(name$, p%, 1))
          If i% = 0 Then i% = 64 ' @
          Do
            Inc i%, Choice(key% = &h80, 1, -1)
            If i% < 0 Then i% = 255
            If i% > 255 Then i% = 0
          Loop Until bits.get%(ALLOWED_CHARS%(i% \ 64), i% Mod 64)
          Do While p% > Len(name$) : Inc name$, " " : Loop
          name$ = Left$(name$, p% - 1) + Chr$(i%) + Mid$(name$, p% + 1)
        Case &h82 ' Left arrow
          If p% > 1 Then Inc p%, -1
        Case &h83 ' Right arrow
          If p% < highscore.MAX_NAME_LEN% Then Inc p%, 1 Else key% = -1
        Case Else
          play_sound% = 0
      End Select
    EndIf

    If play_sound% Then
      start_soundfx(Peek(VarAddr SOUNDFX_SELECT%()))
      play_sound% = 0
    EndIf

  Loop While key% <> -1

  Text x%, y%, Chr$(129), , 1, 1, RGB(Black)
  highscore.values$(idx%) = name$ + ", " + Field$(highscore.values$(idx%), 2)
End Sub

' Gets player input to edit entry in the high-score table.
'
' @param  ctrl$    controller for given player.
' @return          ASCII code of key pressed, or negative of value read
'                  from player's configured controller port.
Function highscore.get_input%(ctrl$)
  highscore.get_input% = Asc(UCase$(Inkey$))
  If highscore.get_input% > 0 Then Exit Function

  Local p%, s$

  ' Note that port initialisation should have already occurred earlier.
  Local k% = Call(ctrl$)
  Select Case k%
    Case NES_A%     : k% = &h83 ' Right arrow
    Case NES_START% : k% = &h0A ' Line feed \n
    Case NES_UP%    : k% = &h80 ' Up arrow
    Case NES_DOWN%  : k% = &h81 ' Down arrow
    Case NES_LEFT%  : k% = &h82 ' Left arrow
    Case NES_RIGHT% : k% = &h83 ' Right arrow
    Case Else       : k% = 0
  End Select
  highscore.get_input% = k%
End Function
' END:       #Include "highscore.inc" ------------------------------------------

Const VERSION$ = "0.9.0"

Select Case MM.Device$
  Case "Colour Maximite 2", "Colour Maximite 2 G2", "MMBasic for Windows"
    Const IS_CMM2% = 1
    MODE 7
    Page Write 1
  Case "PicoMiteVGA"
    Const HIGHSCORE_FILENAME$ = "/high-scores/lazer-cycle.csv"
    Const IS_CMM2% = 0
    MODE 2
  Case Else
    Error "Unsupported device: " + MM.Device$
End Select

Const COLOUR_BLACK%  = RGB(Black)
Const COLOUR_BLUE%   = RGB(Blue)
Const COLOUR_CYAN%   = RGB(Cyan)
Const COLOUR_GREEN%  = RGB(Green)
Const COLOUR_GREY%   = RGB(Grey)
Const COLOUR_RED%    = RGB(Red)
Const COLOUR_YELLOW% = RGB(Yellow)
Const COLOUR_WHITE%  = RGB(White)
Const WIDTH% = MM.HRes \ 2
Const HEIGHT% = (MM.VRes - 20) \ 2
Const X_OFFSET% = MM.HRes \ 2
Const Y_OFFSET% = MM.VRes \ 2
Const NORTH% = 0, EAST% = 1, SOUTH% = 2, WEST% = 3
Const MAX_CYCLE_IDX% = 3
Const SCORE_Y% = 2 * HEIGHT% + 4
Const STATE_OK%    = &b000 ' 0; values 1-3 are "imminent death"
Const STATE_DYING% = &b100 ' 4
Const STATE_DEAD%  = &b101 ' 5

' These would be constants but MMBasic does not support constant arrays
Dim FREQUENCY!(127)
Dim NO_MUSIC%(1)        = (&h0000000000000000, &hFFFFFFFF00000000)
Dim SOUNDFX_NOTHING%(1) = (&hFFFFFFFFFFFFFFFF, &hFFFFFFFFFFFFFFFF)
Dim SOUNDFX_SELECT%(1)  = (&hFFFFFFFF0048443C, &hFFFFFFFFFFFFFFFF)
Dim SOUNDFX_DIE%(3)     = (&h4748494A4B4C4D4E, &h3F40414243444546, &h0038393A3B3C3D3E, &hFFFFFFFFFFFFFFFF)
Dim SOUNDFX_WIPE%(3)    = (&h3F3E3D3C3B3A3938, &h4746454443424140, &h004E4D4C4B4A4948, &hFFFFFFFFFFFFFFFF)
Dim NEXT_DIR%(7)        = (EAST%, NORTH%, WEST%, SOUTH%, EAST%, NORTH%, WEST%, SOUTH%)
Dim SCORE_X%(3)         = (35, 105, 175, 245)
Dim DIRECTIONS%(3)      = (-WIDTH%, 1, WIDTH%, -1)
Dim COMPASS_TO_CTRL%(3) = (NES_UP%, NES_RIGHT%, NES_DOWN%, NES_LEFT%)
Dim CTRL_NAMES$(8)
Dim CTRL_SUBS$(8)

Dim default_ctrl$ = ""
Dim attract_mode% = 1
Dim score%
Dim difficulty% = 1
Dim frame_duration%
Dim next_frame%

' Each cell of the arena takes up 1 byte:
'   bit  0    - occupied by cycle
'   bits 1-2 - index of cycle
'   bits 3-4 - direction cycle was going in when entered cycle
'   bits 5-6 - unused
'   bit  7   - arena wall (other bits will be 0)
Dim arena%(HEIGHT% * WIDTH% \ 8)

Dim cycle.current% ' Current cycle index, set before calling ctrl_xxx() functions.
Dim cycle.is_ai%(MAX_CYCLE_IDX%)
Dim cycle.score%(MAX_CYCLE_IDX%)
Dim cycle.nxt%(MAX_CYCLE_IDX%)
Dim cycle.pos%(MAX_CYCLE_IDX%)
Dim cycle.dir%(MAX_CYCLE_IDX%)
Dim cycle.colour%(MAX_CYCLE_IDX%) = (COLOUR_RED%, COLOUR_YELLOW%, COLOUR_CYAN%, COLOUR_GREEN%)
Dim cycle.ctrl$(MAX_CYCLE_IDX%) Length 32
Dim cycle.ctrl_backup$(MAX_CYCLE_IDX%) Length 32
Dim cycle.state%(MAX_CYCLE_IDX%)

Dim music_start_ptr% = Peek(VarAddr NO_MUSIC%())
Dim music_ptr% = music_start_ptr%
Dim soundfx_flag% = 1
Dim soundfx_ptr% = Peek(VarAddr SOUNDFX_NOTHING%())
Dim num_players%
Dim num_humans%
Dim keys%(31)

Option Break 4
On Key 3, break_handler

init_globals()
read_music()
clear_display()

music_start_ptr% = Peek(VarAddr MUSIC%())

' Music and sound effects are played on SetTick interrupts.
SetTick 200, play_music, 1
SetTick 40, play_soundfx, 2

Do
  If attract_mode% Then
    wipe()
    attract_mode% = Not show_title%(5000)
  EndIf

  If attract_mode% Then
    wipe()
    attract_mode% = Not highscore.show_table%(5000)
  EndIf

  wipe()

  If attract_mode% Then
    init_attract_mode()
  Else
    show_menu()
    wipe()
  EndIf

  init_game()
  draw_arena()

  Select Case game_loop%()
    Case 0
      show_game_over()
      attract_mode% = Not highscore.show_table%(5000)
    Case 1
      attract_mode% = 1
    Case 2
      attract_mode% = 0
    Case Else
      Error "Invalid value for 'attract_mode%'"
  End Select

Loop

End

' Break handler to stop music & fx when Ctrl-C pressed.
Sub break_handler()
  Play Stop
  End
End Sub

' Initialises global variables.
Sub init_globals()
  Local i%
  ' FREQUENCY(0) - rest - 10 Hz, which should be inaudible.
  ' FREQUENCY(1) - C0   - 16.35 Hz
  FREQUENCY!(0) = 10.0
  For i% = 1 To 127
    FREQUENCY!(i%) = 440 * 2^((i% - 58) / 12.0)
  Next

  Restore controller_data
  For i% = 0 To Bound(CTRL_NAMES$(), 1)
    Read CTRL_NAMES$(i%), CTRL_SUBS$(i%)
  Next

  Restore highscore_data
  Local defaults$(Bound(highscore.values$(), 1))
  For i% = 0 To Bound(defaults$(), 1)
    Read defaults$(i%)
  Next
  highscore.load(HIGHSCORE_FILENAME$, defaults$())
End Sub

Sub read_music()
  Restore music_data
  Local count%
  Read count%
  Local num_channels%
  Read num_channels%
  count% = count% \ 8
  Dim MUSIC%(count%)
  Local i%
  For i% = 1 To count%
    Read MUSIC%(i% - 1)
  Next
End Sub

' Displays the title screen for a specified duration or until the user presses
' START/FIRE/SPACE.
'
' @param duration%  duration in milliseconds; if 0 then indefinite.
' @return           1 if the user pressed button/key,
'                   0 if the duration expired.
Function show_title%(duration%)
  Text X_OFFSET%, Y_OFFSET% - 15, "LAZER CYCLE", "CM", 1, 2, COLOUR_WHITE%
  Text X_OFFSET%, Y_OFFSET% + 8, "(c) 2022 Thomas Hugo Williams", "CM", 7, 1, COLOUR_CYAN%
  Text X_OFFSET%, Y_OFFSET% + 20, "www.sockpuppetstudios.com", "CM", 7, 1, COLOUR_CYAN%
  Local text$ = Choice(MM.Device$ = "PicoMiteVGA", "Press START, FIRE or SPACE", "Press SPACE")
  Text X_OFFSET%, Y_OFFSET% + 40, text$, "CM", 1, 1, COLOUR_WHITE%
  If IS_CMM2% Then Page Copy 1 To 0, B
  show_title% = wait%(duration%)
End Function

Sub clear_display()
   Box 0, 0, MM.HRes, MM.VRes, 1, COLOUR_BLACK%, COLOUR_BLACK%
   If IS_CMM2% Then Page Copy 1 To 0, B
End Sub

' Waits a specified duration for the user to press START on a (S)NES gamepad
' connected to Port A, or FIRE on an ATARI joystick connected to Port A or
' SPACE on the keyboard.
'
' @param duration%  duration in milliseconds; if 0 then indefinite.
' @return           1 if the user pressed button/key,
'                   0 if the duration expired.
Function wait%(duration%)
  keyboard_init()
  Local ctrl$ = poll_for_controller$(duration%)
  wait% = ctrl$ <> ""
  default_ctrl$ = Choice(default_ctrl$ = "", ctrl$, default_ctrl$)
End Function

Function str.lpad$(s$, x%)
  str.lpad$ = s$
  If Len(s$) < x% Then str.lpad$ = Space$(x% - Len(s$)) + s$
End Function

Function str.rpad$(s$, x%)
  str.rpad$ = s$
  If Len(s$) < x% Then str.rpad$ = s$ + Space$(x% - Len(s$))
End Function

Sub show_menu()
  Const x% = X_OFFSET% - 100
  Local key%, i%, item% = 0, update% = 1
  Local sounds$(3) = ("MUSIC & FX", "MUSIC ONLY", "FX ONLY   ", "NONE      ")
  Local sound_setting% = Choice(music_start_ptr% = Peek(VarAddr MUSIC%()), 1, 3) - soundfx_flag%

  ' Initialise cycle.ctrl_backup$() on first entry to this subroutine,
  ' subsequent calls use the previously stored state.
  If default_ctrl$ = "" Then default_ctrl$ = "keyboard"
  If cycle.ctrl_backup$(0) = "" Then
    cycle.ctrl_backup$(0) = "ctrl_" + default_ctrl$ + "_a%"
    If cycle.ctrl_backup$(0) = "ctrl_keyboard_a%" Then cycle.ctrl_backup$(0) = "ctrl_keys1%"
    For i% = 1 To MAX_CYCLE_IDX% : cycle.ctrl_backup$(i%) = "ctrl_ai%" : Next
  EndIf

  For i% = 0 To MAX_CYCLE_IDX% : cycle.ctrl$(i%) = cycle.ctrl_backup$(i%) : Next

  ' Text X_OFFSET%, Y_OFFSET% + 75, "Music by Scott Joplin", "CM", 7, 1, COLOUR_CYAN%
  Text X_OFFSET%, Y_OFFSET% + 90, "Game Version " + VERSION$, "CM", 7, 1, COLOUR_CYAN%

  keyboard_init()
  Call default_ctrl$ + "_init", PORT_A%

  Do

    If update% Then
      Text x%, Y_OFFSET% - 95, "START GAME", , 1, 1, COLOUR_WHITE%
      text_for_controller(x%, Y_OFFSET% - 75, 0)
      text_for_controller(x%, Y_OFFSET% - 55, 1)
      text_for_controller(x%, Y_OFFSET% - 35, 2)
      text_for_controller(x%, Y_OFFSET% - 15, 3)
      Text x%, Y_OFFSET% + 5,  "DIFFICULTY: " + Str$(difficulty%), , 1, 1, COLOUR_WHITE%
      Text x%, Y_OFFSET% + 25, "SOUND:      " + sounds$(sound_setting%), , 1, 1, COLOUR_WHITE%
      Text x%, Y_OFFSET% + 45, "QUIT", , 1, 1, COLOUR_WHITE%
      Text x% - 15, Y_OFFSET% - 95 + item% * 20, Chr$(137), , 1, 1, COLOUR_CYAN%
      If IS_CMM2% Then Page Copy 1 To 0, B
      Pause 100
      keyboard_init()
      update% = 0
    EndIf

    key% = Call(default_ctrl$ + "_read%", PORT_A%)
    If key% = 0 Then key% = keyboard_read%(PORT_A%) ' Always respond to the keyboard.
    Select Case key%
      Case NES_UP%
        If item% > 0 Then
          Text x% - 15, Y_OFFSET% - 95 + item% * 20, Chr$(137), , 1, 1, COLOUR_BLACK%
          Inc item%, -1
          update% = 1
        EndIf

      Case NES_DOWN%
        If item% < 7 Then
          Text x% - 15, Y_OFFSET% - 95 + item% * 20, Chr$(137), , 1, 1, COLOUR_BLACK%
          Inc item%
          update% = 1
        EndIf

      Case NES_LEFT%, NES_RIGHT%, NES_A%
        Select Case item%
          Case 0
            If key% = NES_A% Then Exit

          Case 1,2,3,4
            update_controller(item% - 1, Choice(key% = NES_LEFT%, -1, +1))
            update% = 1

          Case 5 ' Difficulty
            Inc difficulty%, Choice(key% = NES_LEFT%, -1, 1)
            If difficulty% < 1 Then difficulty% = 5
            If difficulty% > 5 Then difficulty% = 1
            update% = 1

          Case 6 ' Sound
            Inc sound_setting%, Choice(key% = NES_LEFT%, -1, 1)
            If sound_setting% < 0 Then sound_setting% = 3
            If sound_setting% > 3 Then sound_setting% = 0
            music_start_ptr% = Choice(sound_setting% And &b10, Peek(VarAddr NO_MUSIC%()), Peek(VarAddr MUSIC%()))
            music_ptr% = music_start_ptr%
            soundfx_flag% = Not (sound_setting% And &b01)
            update% = 1

          Case 7 ' Quit
            If key% = NES_A% Then End

        End Select

    End Select

    If update% = 1 Then start_soundfx(Peek(VarAddr SOUNDFX_SELECT%()), 1)

  Loop

  For i% = 0 To MAX_CYCLE_IDX% : cycle.ctrl_backup$(i%) = cycle.ctrl$(i%) : Next
End Sub

Sub text_for_controller(x%, y%, idx%)
  Local txt$ = "PLAYER " + Str$(idx% + 1) + ":   " + get_controller_name$(idx%)
  Text x%, y%, txt$, , 1, 1, cycle.colour%(idx%)
End Sub

Function get_controller_name$(idx%)
  Local i%
  For i% = 0 To Bound(CTRL_SUBS$(), 1)
    If cycle.ctrl$(idx%) = CTRL_SUBS$(i%) Then
      get_controller_name$ = CTRL_NAMES$(i%)
      Do While Len(get_controller_name$) < 12 : Inc get_controller_name$, " " : Loop
      Exit Function
    EndIf
  Next
  Error "Unknown controller: " + cycle.ctrl$(idx%)
End Function

Sub update_controller(idx%, delta%)
  ' Find index of currently selected controller.
  Local i%
  For i% = 0 To Bound(CTRL_SUBS$(), 1)
    If cycle.ctrl$(idx%) = CTRL_SUBS$(i%) Then Exit For
  Next
  If i% = Bound(CTRL_SUBS$(), 1) + 1 Then Error "Unknown controller: " + cycle.ctrl$(idx%)

  Local ok% = 1
  Do
    Inc i%, delta%
    If i% < 0 Then i% = Bound(CTRL_SUBS$(), 1)
    If i% > Bound(CTRL_SUBS$(), 1) Then i% = 0

    ' Check there is no conflict with other player's controller choice.
    Select Case CTRL_SUBS$(i%)
      Case "atari_a_read%", "nes_a_read%"
        ok% = Not has_controller%(idx%, "atari_a_read%")
        ok% = ok% And (Not has_controller%(idx%, "nes_a_read%"))
      Case "atari_b_read%", "nes_b_read%"
        ok% = Not has_controller%(idx%, "atari_b_read%")
        ok% = ok% And (Not has_controller%(idx%, "nes_b_read%"))
      Case "ctrl_keys1%", "ctrl_keys2%"
        ok% = Not has_controller%(idx%, "ctrl_keys1%")
        ok% = ok% And (Not has_controller%(idx%, "ctrl_keys2%"))
      Case "ctrl_keys3%"
        ok% = Not has_controller%(idx%, "ctrl_keys3%")
    End Select
  Loop Until ok%

  cycle.ctrl$(idx%) = CTRL_SUBS$(i%)
End Sub

' Is any player other than idx% using the specified controller ?
Function has_controller%(idx%, ctrl$)
  Local i%
  For i% = 0 To MAX_CYCLE_IDX%
    If i% = idx% Then Continue For
    If cycle.ctrl$(i%) = ctrl$ Then has_controller% = 1
  Next
End Function

Sub init_attract_mode()
  Local i%
  For i% = 0 To MAX_CYCLE_IDX%
    cycle.ctrl$(i%) = "ctrl_ai%"
  Next
  attract_mode% = 1
End Sub

Sub init_game()
  frame_duration% = 3 * (5 + (6 - difficulty%))
  num_players% = 0 ' Incremented later.
  num_humans% = 0
  score% = 0
  Memory Set Peek(VarAddr keys%()), 0, 256

  ' Initialise the arena.
  Local p_arena% = Peek(VarAddr arena%())
  Memory Set p_arena%, 0, HEIGHT% * WIDTH%
  Memory Set p_arena%, 128, WIDTH%
  Memory Set p_arena% + (HEIGHT% - 1) * WIDTH%, 128, WIDTH%
  Local y%
  For y% = 1 To HEIGHT% - 2
    Poke Byte p_arena% + y% * WIDTH%, 128
    Poke Byte p_arena% + (y% + 1) * WIDTH% - 1, 128
  Next

  cycle.dir%(0) = EAST%
  cycle.dir%(1) = WEST%
  cycle.dir%(2) = SOUTH%
  cycle.dir%(3) = NORTH%

  cycle.pos%(0) = WIDTH * (HEIGHT% \ 2) + 5
  cycle.pos%(1) = WIDTH% * (HEIGHT% \ 2) + WIDTH% - 6
  cycle.pos%(2) = 5.5 * WIDTH%
  cycle.pos%(3) = WIDTH% * (HEIGHT% - 6) + WIDTH% \ 2

  Local i%
  For i% = 0 To MAX_CYCLE_IDX%
    cycle.is_ai%(i%) = cycle.ctrl$(i%) = "ctrl_ai%"
    cycle.score%(i%) = 0
    If cycle.ctrl$(i%) = "ctrl_none%" Then
      cycle.pos%(i%) = -1
      cycle.nxt%(i%) = -1
      cycle.state%(i%) = STATE_DEAD%
    Else
      Inc num_players%
      If cycle.ctrl$(i%) <> "ctrl_ai%" Then Inc num_humans%
      cycle.nxt%(i%) = cycle.pos%(i%) + DIRECTIONS%(cycle.dir%(i%))
      Poke Byte p_arena% + cycle.pos%(i%), (cycle.dir%(i%) << 3) + (i% << 1) + 1
      Poke Byte p_arena% + cycle.nxt%(i%), (cycle.dir%(i%) << 3) + (i% << 1) + 1
      cycle.state%(i%) = STATE_OK%
    EndIf
  Next

  ' Initialise game ports.
  For i% = 0 To MAX_CYCLE_IDX%
    Select Case cycle.ctrl$(i%)
      Case "nes_a_read%"   : nes_init_a()
      Case "nes_b_read%"   : nes_init_b()
      Case "atari_a_read%" : atari_init_a()
      Case "atari_b_read%" : atari_init_b()
    End Select
  Next

End Sub

Sub draw_arena()
  Local a%, i%, j%
  For i% = 0 To Bound(arena%(), 1) - 1
    a% = arena%(i%)
    If a% = 0 Then Continue For
    For j% = 0 To 7
      If Peek(Var a%, j%) <> 128 Then Continue For
      Pixel 2 * (((i% * 8) Mod WIDTH%) + j%), 2 * ((i% * 8) \ WIDTH%), COLOUR_GREY%
    Next
  Next
End Sub

' @return  0 - normal game over
'          1 - attract mode game over
'          2 - attract mode interrupted
Function game_loop%()
  Local ctrl$, i%
  game_loop% = attract_mode%
  next_frame% = Timer + frame_duration%

  ' num_players% = 0
  ' cycle.score%(0) = 3175
  ' cycle.score%(1) = 2175
  ' cycle.score%(2) = 1175
  ' cycle.score%(3) = 975
  ' score% = 3175

  keyboard_init()
  If Not attract_mode% Then On Key on_key()

  Do While num_players% > 0
    Inc score%, 1
    If score% Mod 5 = 0 Then draw_score()

    ' When dying the cycle trail deletes at twice the rate.
    For i% = 0 To MAX_CYCLE_IDX%
      If cycle.state%(i%) = STATE_DYING% Then cycle.dying(i%)
    Next

    ' Draw cycles.
    For i% = 0 To MAX_CYCLE_IDX%
      If Not (cycle.state%(i%) And &b11) Then cycle.draw(i%)
    Next

    If IS_CMM2% Then Page Copy 1 To 0, I

    ' Move cycles.
    For i% = 0 To MAX_CYCLE_IDX%
      If Not (cycle.state%(i%) And &b11) Then cycle.pos%(i%) = cycle.nxt%(i%)
    Next

    ' Determine changes of direction and check for collisions.
    For i% = 0 To MAX_CYCLE_IDX%
      cycle.current% = i%
      Select Case Call(cycle.ctrl$(i%))
        Case NES_UP%:    cycle.dir%(i%) = NORTH%
        Case NES_DOWN%:  cycle.dir%(i%) = SOUTH%
        Case NES_LEFT%:  cycle.dir%(i%) = WEST%
        Case NES_RIGHT%: cycle.dir%(i%) = EAST%
      End Select
      cycle.nxt%(i%) = cycle.pos%(i%) + DIRECTIONS%(cycle.dir%(i%))
      If cycle.state%(i%) <> STATE_DEAD% Then cycle.check_collision(i%)
    Next

    ' Wait for next frame.
    Do While Timer < next_frame% : Loop
    Inc next_frame%, frame_duration%

    If Not attract_mode% Then Continue Do

    ' Check for "attract mode" being interrupted.
    ctrl$ = poll_for_controller$(1)
    If ctrl$ <> "" Then
      num_players% = 0
      game_loop% = 2
    EndIf
  Loop

  ' Ensure display updated at end of loop.
  If IS_CMM2% Then Page Copy 1 To 0, B

  ' Wait for any sound effect to complete.
  Do While Peek(Byte soundfx_ptr%) <> &hFF : Loop

  If Not attract_mode% Then On Key 0
End Function

Sub draw_score()
  Local i%, s$ = Str$(score%, 5, 0, "0")
  For i% = 0 To MAX_CYCLE_IDX%
    If cycle.state%(i%) < STATE_DYING% Then
      Text SCORE_X%(i%), SCORE_Y%, s$, , 1, 1, cycle.colour%(i%)
    EndIf
  Next
End Sub

Sub wipe()
  Local y%
  start_soundfx(Peek(VarAddr SOUNDFX_WIPE%()), 1)
  For y% = 0 To MM.VRes \ 2 Step 5
     Box WIDTH% - y% * 1.2, MM.VRes \ 2 - y%, 2.4 * y%, 2 * y%, 5, COLOUR_CYAN%, COLOUR_BLACK%
     If IS_CMM2% Then Page Copy 1 To 0, B
     Pause 30
  Next
  clear_display()
End Sub

' Draw cycle if STATE_OK% or STATE_DYING%.
Sub cycle.draw(idx%)
  Local p% = cycle.pos%(idx%), n% = cycle.nxt%(idx%)
  Line 2*(p% Mod WIDTH%), 2*(p%\WIDTH%), 2*(n% Mod WIDTH%), 2*(n%\WIDTH%), 1, cycle.colour%(idx%) * (cycle.state%(idx%) <> STATE_DYING%)
End Sub

Sub cycle.dying(idx%)
  cycle.draw(idx%)
  cycle.pos%(idx%) = cycle.nxt%(idx%) ' Move
  cycle.current% = idx%
  Select Case ctrl_die%()
    Case NES_UP%:    cycle.dir%(idx%) = NORTH%
    Case NES_DOWN%:  cycle.dir%(idx%) = SOUTH%
    Case NES_LEFT%:  cycle.dir%(idx%) = WEST%
    Case NES_RIGHT%: cycle.dir%(idx%) = EAST%
  End Select
  cycle.nxt%(idx%) = cycle.pos%(idx%) + DIRECTIONS%(cycle.dir%(idx%))
  cycle.check_collision(idx%)
End Sub

Sub show_game_over()
  ' Sort scores and then round down to nearest 5.
  Local dummy%, i%, idx%(MAX_CYCLE_IDX%), j%, k%, winner%
  Sort cycle.score%(), idx%(), 1
  For i% = 0 To MAX_CYCLE_IDX%
    Do While (cycle.score%(i%) Mod 5) <> 0 : Inc cycle.score%(i%), -1 : Loop
  Next
  winner% = idx%(0)

  Local txt$ = "PLAYER " + Str$(winner% + 1) + " WINS"
  Text X_OFFSET%, Y_OFFSET% - 25, txt$, "CM", 1, 2, cycle.colour%(winner%)
  Text X_OFFSET%, Y_OFFSET% + 5, "SCORE: " + Str$(cycle.score%(0)), "CM", 1, 2, cycle.colour%(winner%)
  If IS_CMM2% Then Page Copy 1 To 0, B
  dummy% = wait%(5000)

  wipe()

  ' Insert into high-score table.
  Local new_highscore%, player%
  For i% = 0 To MAX_CYCLE_IDX%
    player% = idx%(i%)
    If cycle.is_ai%(player%) Then Continue For
    For j% = 0 To Bound(highscore.values$())
      If cycle.score%(i%) > Val(Field$(highscore.values$(j%), 2)) Then
        For k% = Bound(highscore.values$(), 1) To j% Step -1
          If k% <> 0 Then highscore.values$(k%) = highscore.values$(k% - 1)
        Next
        highscore.values$(j%) = ", " + Str$(cycle.score%(i%))
        highscore.edit(j%, cycle.colour%(player%), cycle.ctrl_backup$(player%))
        new_highscore% = 1
        Exit For
      EndIf
    Next
  Next
  If new_highscore% Then highscore.save(HIGHSCORE_FILENAME$)
End Sub

Function ctrl_ai%()
  Local idx% = cycle.current%
  Local d% = cycle.dir%(idx%)

  ' Random element.
  Local i% = Int(500 * Rnd)
  If i% < 4 Then d% = NEXT_DIR%(i% + idx%)

  ' Avoid collisions.
  Local nxt%
  For i% = 0 To MAX_CYCLE_IDX%
    nxt% = cycle.pos%(idx%) + DIRECTIONS%(d%)
    If Not Peek(Var arena%(), nxt%)) Then Exit For
    d% = NEXT_DIR%(i% + idx%)
  Next

  ctrl_ai% = COMPASS_TO_CTRL%(d%)
End Function

Function ctrl_die%()
  ctrl_die% = COMPASS_TO_CTRL%(((Peek(Var arena%(), cycle.pos%(cycle.current%)) >> 3) + 2) Mod 4)
End Function

Function ctrl_keys1%()
  If Peek(Var keys%(), 65) Then      ' A
    ctrl_keys1% = NES_UP%
    Poke Var keys%(), 65, 0
  ElseIf Peek(Var keys%(), 90) Then  ' Z
    ctrl_keys1% = NES_DOWN%
    Poke Var keys%(), 90, 0
  ElseIf Peek(Var keys%(), 46) Then   ' full-stop
    ctrl_keys1% = NES_RIGHT%
    Poke Var keys%(), 46, 0
  ElseIf Peek(Var keys%(), 44) Then   ' comma
    ctrl_keys1% = NES_LEFT%
    Poke Var keys%(), 44, 0
  EndIf
End Function

Function ctrl_keys2%()
  If Peek(Var keys%(), 65) Then      ' A
    ctrl_keys2% = NES_UP%
    Poke Var keys%(), 65, 0
  ElseIf Peek(Var keys%(), 90) Then  ' Z
    ctrl_keys2% = NES_DOWN%
    Poke Var keys%(), 90, 0
  ElseIf Peek(Var keys%(), 67) Then  ' C
    ctrl_keys2% = NES_RIGHT%
    Poke Var keys%(), 67, 0
  ElseIf Peek(Var keys%(), 88) Then  ' X
    ctrl_keys2% = NES_LEFT%
    Poke Var keys%(), 88, 0
  EndIf
End Function

Function ctrl_keys3%()
  If Peek(Var keys%(), 73) Then      ' I
    ctrl_keys3% = NES_UP%
    Poke Var keys%(), 73, 0
  ElseIf Peek(Var keys%(), 75) Then  ' K
    ctrl_keys3% = NES_DOWN%
    Poke Var keys%(), 75, 0
  ElseIf Peek(Var keys%(), 80) Then  ' P
    ctrl_keys3% = NES_RIGHT%
    Poke Var keys%(), 80, 0
  ElseIf Peek(Var keys%(), 79) Then  ' O
    ctrl_keys3% = NES_LEFT%
    Poke Var keys%(), 79, 0
  EndIf
End Function

Function ctrl_none%()
  ' Do nothing.
End Function

Sub cycle.check_collision(idx%)
  ' Handle dying.
  If cycle.state%(idx%) = STATE_DYING% Then
    Poke Var arena%(), cycle.pos%(idx%), 0
    Local mask% = (idx% << 1) + 1
    If (Peek(Var arena%(), cycle.nxt%(idx%)) And mask%) <> mask% Then
      cycle.ctrl$(idx%) = "ctrl_none%"
      cycle.state%(idx%) = STATE_DEAD%
      cycle.pos%(idx%) = -1
    EndIf
    Exit Sub
  EndIf

  ' No collision occurred.
  If Not Peek(Var arena%(), cycle.nxt%(idx%)) Then
    Poke Var arena%(), cycle.nxt%(idx%), (cycle.dir%(idx%) << 3) + (idx% << 1) + 1
    cycle.state%(idx%) = STATE_OK%
    Exit Sub
  EndIf

  ' Collision occured - the player has a couple of frames to change direction.
  Inc cycle.state%(idx%)
  If cycle.state%(idx%) < STATE_DYING% Then Exit Sub

  ' Time to die.
  Inc num_players%, -1
  If cycle.ctrl$(idx%) <> "ctrl_ai%" Then
    Inc num_humans%, -1
    If num_humans% = 0 Then
      On Key 0
      attract_mode% = 1
    EndIf
  EndIf
  cycle.ctrl$(idx%) = "ctrl_die%"
  cycle.nxt%(idx%) = cycle.pos%(idx%)
  cycle.score%(idx%) = score%
  start_soundfx(Peek(VarAddr SOUNDFX_DIE%()), 0)
End Sub

Sub on_key()
  Local ch$ = Inkey$
  If ch$ <> "" Then Poke Var keys%(), Asc(UCase$(ch$)), 1
End Sub

' Called from interrupt to play next note of music.
Sub play_music()
  Local n% = Peek(Byte music_ptr%)
  If n% = 255 Then music_ptr% = music_start_ptr% : Exit Sub
  Play Sound 1, B, S, FREQUENCY!(n%), 15
  Play Sound 2, B, S, FREQUENCY!(Peek(Byte music_ptr% + 1)), 15
  Play Sound 3, B, S, FREQUENCY!(Peek(Byte music_ptr% + 2)), 15
  Inc music_ptr%, 3
End Sub

' Start a new sound effect.
Sub start_soundfx(ptr%, wait_%)
  If Not soundfx_flag% Then Exit Sub

  ' Wait for current sound effect to end.
  If wait_% Then Do While Peek(Byte soundfx_ptr%) <> &hFF : Loop

  soundfx_ptr% = ptr%

  ' Wait for first note of new sound effect to play.
  Do While soundfx_ptr% = ptr% : Loop
End Sub

' Called from interrupt to play next note of current sound effect.
Sub play_soundfx()
  Local n% = Peek(Byte soundfx_ptr%)
  If n% = 255 Then Exit Sub
  Play Sound 4, B, s, FREQUENCY!(n%), (n% <> 0) * 25
  Inc soundfx_ptr%
End Sub

' Gets bit i% of x%.
Function bits.get%(x%, i%)
  If i% < 0 Or i% > 63 Then Error "i% out of 0 .. 63 range"
  bits.get% = (x% And 1 << i%) <> 0
End Function

music_data:

Data 792   ' Number of bytes of music data.
Data 3     ' Number of channels.
Data &h3135000034000033, &h3500253D00313D00, &h00283D00283D0025, &h2A3D00293D002935
Data &h3D002C3D002A3D00, &h002E3D002E00002C, &h314100304000303F, &h4100293F00313D00
Data &h002A3C002A410029, &h253D002C3F002C3F, &h3D002C3D00253D00, &h00313D00313D002C
Data &h3135000034003133, &h3500253D00313D00, &h00283D00283D0025, &h2A3D00293D002935
Data &h3D002C3D002A3D00, &h002E3D002E3D002C, &h273700263800263A, &h41002B3D00273A00
Data &h002E3F002E41002B, &h2C3F00273A00273D, &h3F002A3F002C3F00, &h00293F00293F002A
Data &h2535002734002733, &h3500313D00253D00, &h00283D00283D0031, &h2A3D00293D002935
Data &h3D002C3D002A3D00, &h002E3D002E00002C, &h314100304000303F, &h4100293F00313D00
Data &h002A3C002A410029, &h313D002C3F002C3F, &h3D002C3D00313D00, &h00250000253D002C
Data &h3D4100253F00253D, &h41313D3F00003D31, &h00003D2F3B410000, &h3A4100003D2F3B3F
Data &h412E3A3F00003D2E, &h00003D2D39410000, &h384100003D2D393F, &h412C383F00003D2C
Data &h2C383C2C38410000, &h003D00003F00003F, &h3D202C3D00003D00, &h00003525313D202C
Data &h3538000037000036, &h3800363A31353831, &h2C00352C35380035, &h3538000037000036
Data &h3800363A31353831, &h2C00412C35380035, &h003A00003800003D, &h3F2A003D2A003C2A
Data &h2C003F2C00412A00, &h00382C003F2C003D, &h382C003825003825, &h3100353100382C00
Data &h3538000037000036, &h3800363A31353831, &h2C00352C35380035, &h3538000037000036
Data &h3800363A31353831, &h2C00382C35380035, &h003C32003B32003A, &h3C33000033003C33
Data &h27003A27003C3300, &h0038270033270037, &h382C00380000382C, &h2E00352E00380000
Data &h3538300037300036, &h3800363A31353831, &h2C00352C35380035, &h3538000037000036
Data &h3800363A31353831, &h2C00412C35380035, &h003A00003800003D, &h3F2A003D2A003C2A
Data &h2C003F2C00412A00, &h003D2C003F2C003D, &h3D2C003D25003D25, &h31003831003D2C00
Data &h363D000038000037, &h3D2A363A00003D2A, &h00003A2B373D0000, &h383800003A2B373D
Data &h442C384100003D2C, &h0000412935440000, &h363A00003829353D, &h3D2A363D00003A2A
Data &h00003F2C38410000, &h3D3D00003F2C383F, &h3D2C383D313D3D31, &h25313D25313D2C38
Data &hFFFF000000000000, &hFFFFFFFFFFFFFFFF, &hFFFFFFFFFFFFFFFF

controller_data:

Data "KEYS: AZ,.", "ctrl_keys1%"
Data "KEYS: AZXC", "ctrl_keys2%"
Data "KEYS: IKOP", "ctrl_keys3%"
Data "GAMEPAD A",  "nes_a_read%"
Data "GAMEPAD B",  "nes_b_read%"
Data "JOYSTICK A", "atari_a_read%"
Data "JOYSTICK B", "atari_b_read%"
Data "AI",         "ctrl_ai%"
Data "NONE",       "ctrl_none%"

highscore_data:

Data "TOM, 2000"
Data "MICKEY, 1500"
Data "MIKE, 1250"
Data "PETER, 1000"
Data "DAVEY, 800"
Data "JOHN, 600"
Data "PAUL, 400"
Data "GEORGE, 200"
Data "RINGO, 100"
Data "MOOSE, 50"
